@@ -3,7 +3,9 @@ import logging
 
 from aiohttp.client_exceptions import ClientError
 from rest_framework.viewsets import ViewSet
+from rest_framework.renderers import BrowsableAPIRenderer, JSONRenderer, TemplateHTMLRenderer
 from rest_framework.response import Response
+from rest_framework.exceptions import NotAcceptable
 from django.core.exceptions import ObjectDoesNotExist
 from django.shortcuts import redirect
 from datetime import datetime, timezone, timedelta
@@ -13,6 +15,7 @@ from django.db import transaction
 from django.db.utils import DatabaseError
 from django.http.response import (
     Http404,
+    HttpResponseNotFound,
     HttpResponseForbidden,
     HttpResponseBadRequest,
     StreamingHttpResponse,
@@ -43,7 +46,9 @@ from pulp_python.app.pypi.serializers import (
 )
 from pulp_python.app.utils import (
     write_simple_index,
+    write_simple_index_json,
     write_simple_detail,
+    write_simple_detail_json,
     python_content_to_json,
     PYPI_LAST_SERIAL,
     PYPI_SERIAL_CONSTANT,
@@ -57,6 +62,17 @@ log = logging.getLogger(__name__)
 ORIGIN_HOST = settings.CONTENT_ORIGIN if settings.CONTENT_ORIGIN else settings.PYPI_API_HOSTNAME
 BASE_CONTENT_URL = urljoin(ORIGIN_HOST, settings.CONTENT_PATH_PREFIX)
 
+PYPI_SIMPLE_V1_HTML = "application/vnd.pypi.simple.v1+html"
+PYPI_SIMPLE_V1_JSON = "application/vnd.pypi.simple.v1+json"
+
+
+class PyPISimpleHTMLRenderer(TemplateHTMLRenderer):
+    media_type = PYPI_SIMPLE_V1_HTML
+
+
+class PyPISimpleJSONRenderer(JSONRenderer):
+    media_type = PYPI_SIMPLE_V1_JSON
+
 
 class PyPIMixin:
     """Mixin to get index specific info."""
@@ -162,10 +178,15 @@ class PackageUploadMixin(PyPIMixin):
         if settings.PYTHON_GROUP_UPLOADS:
             return self.upload_package_group(repo, artifact, filename, request.session)
 
-        result = dispatch(tasks.upload, exclusive_resources=[artifact, repo],
-                          kwargs={"artifact_sha256": artifact.sha256,
-                                  "filename": filename,
-                                  "repository_pk": str(repo.pk)})
+        result = dispatch(
+            tasks.upload,
+            exclusive_resources=[artifact, repo],
+            kwargs={
+                "artifact_sha256": artifact.sha256,
+                "filename": filename,
+                "repository_pk": str(repo.pk),
+            },
+        )
         return OperationPostponedResponse(result, request)
 
     def upload_package_group(self, repo, artifact, filename, session):
@@ -179,10 +200,10 @@ class PackageUploadMixin(PyPIMixin):
             try:
                 with transaction.atomic():
                     sq.first()
-                    current_start = datetime.fromisoformat(session['start'])
+                    current_start = datetime.fromisoformat(session["start"])
                     if current_start >= datetime.now(tz=timezone.utc):
-                        session['artifacts'].append((str(artifact.sha256), filename))
-                        session['start'] = str(start_time)
+                        session["artifacts"].append((str(artifact.sha256), filename))
+                        session["start"] = str(start_time)
                         session.modified = False
                         session.save()
                     else:
@@ -195,8 +216,8 @@ class PackageUploadMixin(PyPIMixin):
 
     def create_group_upload_task(self, cur_session, repository, artifact, filename, start_time):
         """Creates the actual task that adds the packages to the index."""
-        cur_session['start'] = str(start_time)
-        cur_session['artifacts'] = [(str(artifact.sha256), filename)]
+        cur_session["start"] = str(start_time)
+        cur_session["artifacts"] = [(str(artifact.sha256), filename)]
         cur_session.modified = False
         cur_session.save()
         task = dispatch(
@@ -230,76 +251,127 @@ class SimpleView(PackageUploadMixin, ViewSet):
         ],
     }
 
+    def perform_content_negotiation(self, request, force=False):
+        """
+        Uses standard content negotiation, defaulting to HTML if no acceptable renderer is found.
+        """
+        try:
+            return super().perform_content_negotiation(request, force)
+        except NotAcceptable:
+            return TemplateHTMLRenderer(), TemplateHTMLRenderer.media_type  # text/html
+
+    def get_renderers(self):
+        """
+        Uses custom renderers for PyPI Simple API endpoints, defaulting to standard ones.
+        """
+        if self.action in ["list", "retrieve"]:
+            # Ordered by priority if multiple content types are present
+            return [TemplateHTMLRenderer(), PyPISimpleHTMLRenderer(), PyPISimpleJSONRenderer()]
+        else:
+            return [JSONRenderer(), BrowsableAPIRenderer()]
+
     @extend_schema(summary="Get index simple page")
     def list(self, request, path):
         """Gets the simple api html page for the index."""
         repo_version, content = self.get_rvc()
         if self.should_redirect(repo_version=repo_version):
-            return redirect(urljoin(self.base_content_url, f'{path}/simple/'))
-        names = content.order_by('name').values_list('name', flat=True).distinct().iterator()
-        return StreamingHttpResponse(write_simple_index(names, streamed=True))
+            return redirect(urljoin(self.base_content_url, f"{path}/simple/"))
+        names = content.order_by("name").values_list("name", flat=True).distinct().iterator()
+        media_type = request.accepted_renderer.media_type
+        headers = {"X-PyPI-Last-Serial": str(PYPI_SERIAL_CONSTANT)}
+
+        if media_type == PYPI_SIMPLE_V1_JSON:
+            index_data = write_simple_index_json(names)
+            return Response(index_data, headers=headers)
+        else:
+            index_data = write_simple_index(names, streamed=True)
+            kwargs = {"content_type": media_type, "headers": headers}
+            return StreamingHttpResponse(index_data, **kwargs)
 
     def pull_through_package_simple(self, package, path, remote):
         """Gets the package's simple page from remote."""
+
         def parse_package(release_package):
             parsed = urlparse(release_package.url)
             stripped_url = urlunsplit(chain(parsed[:3], ("", "")))
-            redirect_path = f'{path}/{release_package.filename}?redirect={stripped_url}'
+            redirect_path = f"{path}/{release_package.filename}?redirect={stripped_url}"
             d_url = urljoin(self.base_content_url, redirect_path)
-            return release_package.filename, d_url, release_package.digests.get("sha256", "")
+            return {
+                "filename": release_package.filename,
+                "url": d_url,
+                "sha256": release_package.digests.get("sha256", ""),
+                "requires_python": release_package.requires_python,
+                "metadata_sha256": (release_package.metadata_digests or {}).get("sha256"),
+            }
 
         rfilter = get_remote_package_filter(remote)
         if not rfilter.filter_project(package):
-            raise Http404(f"{package} does not exist.")
+            return {}
 
-        url = remote.get_remote_artifact_url(f'simple/{package}/')
+        url = remote.get_remote_artifact_url(f"simple/{package}/")
         remote.headers = remote.headers or []
         remote.headers.append({"Accept": ACCEPT_JSON_PREFERRED})
         downloader = remote.get_downloader(url=url, max_retries=1)
         try:
             d = downloader.fetch()
-        except ClientError:
-            return HttpResponse(f"Failed to fetch {package} from {remote.url}.", status=502)
-        except TimeoutException:
-            return HttpResponse(f"{remote.url} timed out while fetching {package}.", status=504)
+        except (ClientError, TimeoutException):
+            log.info(f"Failed to fetch {package} simple page from {remote.url}")
+            return {}
 
-        if d.headers["content-type"] == "application/vnd.pypi.simple.v1+json":
+        if d.headers["content-type"] == PYPI_SIMPLE_V1_JSON:
             page = ProjectPage.from_json_data(json.load(open(d.path, "rb")), base_url=url)
         else:
             page = ProjectPage.from_html(package, open(d.path, "rb").read(), base_url=url)
-        packages = [
-            parse_package(p) for p in page.packages if rfilter.filter_release(package, p.version)
-        ]
-        return HttpResponse(write_simple_detail(package, packages))
+        return {
+            p.filename: parse_package(p)
+            for p in page.packages
+            if rfilter.filter_release(package, p.version)
+        }
 
     @extend_schema(operation_id="pypi_simple_package_read", summary="Get package simple page")
     def retrieve(self, request, path, package):
-        """Retrieves the simple api html page for a package."""
+        """Retrieves the simple api html/json page for a package."""
+        media_type = request.accepted_renderer.media_type
+
         repo_ver, content = self.get_rvc()
         # Should I redirect if the normalized name is different?
         normalized = canonicalize_name(package)
+        releases = {}
         if self.distribution.remote:
-            return self.pull_through_package_simple(normalized, path, self.distribution.remote)
-        if self.should_redirect(repo_version=repo_ver):
-            return redirect(urljoin(self.base_content_url, f'{path}/simple/{normalized}/'))
-        packages = (
-            content.filter(name__normalize=normalized)
-            .values_list('filename', 'sha256', 'name')
-            .iterator()
-        )
-        try:
-            present = next(packages)
-        except StopIteration:
-            raise Http404(f"{normalized} does not exist.")
-        else:
-            packages = chain([present], packages)
-            name = present[2]
-        releases = ((f, urljoin(self.base_content_url, f'{path}/{f}'), d) for f, d, _ in packages)
-        return StreamingHttpResponse(write_simple_detail(name, releases, streamed=True))
+            releases = self.pull_through_package_simple(normalized, path, self.distribution.remote)
+        elif self.should_redirect(repo_version=repo_ver):
+            return redirect(urljoin(self.base_content_url, f"{path}/simple/{normalized}/"))
+        if content:
+            packages = content.filter(name__normalize=normalized).values(
+                "filename", "sha256", "metadata_sha256", "requires_python"
+            )
+            local_releases = {
+                p["filename"]: {
+                    **p,
+                    "url": urljoin(self.base_content_url, f"{path}/{p['filename']}"),
+                }
+                for p in packages
+            }
+            releases.update(local_releases)
+        if not releases:
+            return HttpResponseNotFound(f"{normalized} does not exist.")
 
-    @extend_schema(request=PackageUploadSerializer,
-                   responses={200: PackageUploadTaskSerializer},
-                   summary="Upload a package")
+        media_type = request.accepted_renderer.media_type
+        headers = {"X-PyPI-Last-Serial": str(PYPI_SERIAL_CONSTANT)}
+
+        if media_type == PYPI_SIMPLE_V1_JSON:
+            detail_data = write_simple_detail_json(normalized, releases.values())
+            return Response(detail_data, headers=headers)
+        else:
+            detail_data = write_simple_detail(normalized, releases.values())
+            kwargs = {"content_type": media_type, "headers": headers}
+            return HttpResponse(detail_data, **kwargs)
+
+    @extend_schema(
+        request=PackageUploadSerializer,
+        responses={200: PackageUploadTaskSerializer},
+        summary="Upload a package",
+    )
     def create(self, request, path):
         """
         Upload package to the index.
@@ -324,9 +396,11 @@ class MetadataView(PyPIMixin, ViewSet):
         ],
     }
 
-    @extend_schema(tags=["Pypi: Metadata"],
-                   responses={200: PackageMetadataSerializer},
-                   summary="Get package metadata")
+    @extend_schema(
+        tags=["Pypi: Metadata"],
+        responses={200: PackageMetadataSerializer},
+        summary="Get package metadata",
+    )
     def retrieve(self, request, path, meta):
         """
         Retrieves the package's core-metadata specified by
@@ -372,8 +446,7 @@ class PyPIView(PyPIMixin, ViewSet):
         ],
     }
 
-    @extend_schema(responses={200: SummarySerializer},
-                   summary="Get index summary")
+    @extend_schema(responses={200: SummarySerializer}, summary="Get index summary")
     def retrieve(self, request, path):
         """Gets package summary stats of index."""
         repo_ver, content = self.get_rvc()
@@ -399,9 +472,11 @@ class UploadView(PackageUploadMixin, ViewSet):
         ],
     }
 
-    @extend_schema(request=PackageUploadSerializer,
-                   responses={200: PackageUploadTaskSerializer},
-                   summary="Upload a package")
+    @extend_schema(
+        request=PackageUploadSerializer,
+        responses={200: PackageUploadTaskSerializer},
+        summary="Upload a package",
+    )
     def create(self, request, path):
         """
         Upload package to the index.
